@@ -24,13 +24,12 @@ import torch
 import torch.nn as nn
 import torchvision.transforms as transforms
 from PIL import Image
-from torch.utils.data import DataLoader
-from torch.utils.data import Dataset
-from torchvision.utils import make_grid
-from torchvision.utils import save_image
+from torch.utils.data import DataLoader, Dataset
+from torchvision.utils import make_grid, save_image
 
 from labml import lab, tracker, experiment, monit, configs
 from labml.configs import BaseConfigs
+from labml.utils.pytorch import get_modules
 from labml_helpers.device import DeviceConfigs
 from labml_helpers.module import Module
 
@@ -40,11 +39,8 @@ class GeneratorResNet(Module):
     The generator is a residual network.
     """
 
-    def __init__(self, input_shape: Tuple[int, int, int], n_residual_blocks: int):
+    def __init__(self, input_channels: int, n_residual_blocks: int):
         super().__init__()
-        # The number of channels in the input image, which is 3 for RGB images.
-        channels = input_shape[0]
-
         # This first block runs a $7\times7$ convolution and maps the image to
         # a feature map.
         # The output feature map has same height and width because we have
@@ -54,7 +50,7 @@ class GeneratorResNet(Module):
         # `inplace=True` in `ReLU` saves a little bit of memory.
         out_features = 64
         layers = [
-            nn.Conv2d(channels, out_features, kernel_size=7, padding=3, padding_mode='reflect'),
+            nn.Conv2d(input_channels, out_features, kernel_size=7, padding=3, padding_mode='reflect'),
             nn.InstanceNorm2d(out_features),
             nn.ReLU(inplace=True),
         ]
@@ -89,7 +85,7 @@ class GeneratorResNet(Module):
             in_features = out_features
 
         # Finally we map the feature map to an RGB image
-        layers += [nn.Conv2d(out_features, channels, 7, padding=3, padding_mode='reflect'), nn.Tanh()]
+        layers += [nn.Conv2d(out_features, input_channels, 7, padding=3, padding_mode='reflect'), nn.Tanh()]
 
         # Create a sequential module with the layers
         self.layers = nn.Sequential(*layers)
@@ -324,6 +320,11 @@ class Configs(BaseConfigs):
             # Arrange images along y-axis
             image_grid = torch.cat((data_x, gen_y, data_y, gen_x), 1)
 
+        # Create folder to store sampled images
+        images_path = Path(f'images/{self.dataset_name}')
+        if not images_path.exists():
+            images_path.mkdir(parents=True)
+
         # Save grid
         save_image(image_grid, f"images/{self.dataset_name}/{n}.png", normalize=False)
 
@@ -453,6 +454,9 @@ class Configs(BaseConfigs):
                 # Save images at intervals
                 batches_done = epoch * len(self.dataloader) + i
                 if batches_done % self.sample_interval == 0:
+                    # Save models when sampling images
+                    experiment.save_checkpoint()
+                    # Sample images
                     self.sample_images(batches_done)
 
             # Update learning rates
@@ -548,8 +552,8 @@ def setup_models(self: Configs):
     input_shape = (self.img_channels, self.img_height, self.img_width)
 
     # Create the models
-    self.generator_xy = GeneratorResNet(input_shape, self.n_residual_blocks).to(self.device)
-    self.generator_yx = GeneratorResNet(input_shape, self.n_residual_blocks).to(self.device)
+    self.generator_xy = GeneratorResNet(self.img_channels, self.n_residual_blocks).to(self.device)
+    self.generator_yx = GeneratorResNet(self.img_channels, self.n_residual_blocks).to(self.device)
     self.discriminator_x = Discriminator(input_shape).to(self.device)
     self.discriminator_y = Discriminator(input_shape).to(self.device)
 
@@ -606,13 +610,108 @@ def setup_dataloader(self: Configs):
     )
 
 
-def main():
+def train():
+    """
+    ## Train Cycle GAN
+    """
+    # Create configurations
     conf = Configs()
+    # Create an experiment
     experiment.create(name='cycle_gan')
-    experiment.configs(conf, 'run')
+    # Calculate configurations.
+    # It will calculate `conf.run` and all other configs required by it.
+    experiment.configs(conf, {'dataset_name': 'summer2winter_yosemite'}, 'run')
+    # Register models for saving and loading.
+    # `get_modules` gives a dictionary of `nn.Modules` in `conf`.
+    # You can also specify a custom dictionary of models.
+    experiment.add_pytorch_models(get_modules(conf))
+    # Start and watch the experiment
     with experiment.start():
+        # Run the training
         conf.run()
 
 
+def plot_image(img: torch.Tensor):
+    """
+    ### Plots an image with matplotlib
+    """
+    from matplotlib import pyplot as plt
+
+    # Get min and max values of the image for normalization
+    img_min, img_max = img.min(), img.max()
+    # Scale image values to be [0...1]
+    img = (img - img_min) / (img_max - img_min + 1e-5)
+    # We have to change the order of dimensions to HWC.
+    img = img.permute(1, 2, 0)
+    # Show image
+    plt.imshow(img)
+    plt.show()
+
+
+def evaluate():
+    """
+    ## Evaluate trained Cycle GAN
+    """
+    # Set the run uuid from the training run
+    trained_run_uuid = 'f73c1164184711eb9190b74249275441'
+    # Create configs object
+    conf = Configs()
+    # Create experiment
+    experiment.create(name='cycle_gan_inference')
+    # Load hyper parameters set for training
+    conf_dict = experiment.load_configs(trained_run_uuid)
+    # Calculate configurations. We specify the generators `'generator_xy', 'generator_yx'`
+    # so that it only loads those and their dependencies.
+    # Configs like `device` and `img_channels` will be calculated since these are required by
+    # `generator_xy` and `generator_yx`.
+    #
+    # If you want other parameters like `dataset_name` you should specify them here.
+    # If you specify nothing all the configurations will be calculated including data loaders.
+    # Calculation of configurations and their dependencies will happen when you call `experiment.start`
+    experiment.configs(conf, conf_dict, 'generator_xy', 'generator_yx')
+    # Register models for saving and loading.
+    # `get_modules` gives a dictionary of `nn.Modules` in `conf`.
+    # You can also specify a custom dictionary of models.
+    experiment.add_pytorch_models(get_modules(conf))
+    # Specify which run to load from.
+    # Loading will actually happen when you call `experiment.start`
+    experiment.load(trained_run_uuid)
+
+    # Start the experiment
+    with experiment.start():
+        # Load your own data, here we try test set.
+        # I was trying with yosemite photos, they look awesome.
+        # You can use `conf.dataset_name`, if you specified `dataset_name` as something you wanted to be calculated
+        # in the call to `experiment.configs`
+        images_path = lab.get_data_path() / 'cycle_gan' / 'summer2winter_yosemite'
+
+        # Image transformations
+        transforms_ = [
+            transforms.ToTensor(),
+            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
+        ]
+
+        # Load dataset
+        dataset = ImageDataset(images_path, transforms_, True, 'train')
+        # Get an images from dataset
+        x_image = dataset[10]['x']
+        # Display the image
+        plot_image(x_image)
+
+        # Evaluation mode
+        conf.generator_xy.eval()
+        conf.generator_yx.eval()
+
+        # We dont need gradients
+        with torch.no_grad():
+            # Add batch dimension and move to the device we use
+            data = x_image.unsqueeze(0).to(conf.device)
+            generated_y = conf.generator_xy(data)
+
+        # Display the generated image.
+        plot_image(generated_y[0].cpu())
+
+
 if __name__ == '__main__':
-    main()
+    train()
+    # evaluate()
